---
examples:
 - name: basic
   label: Button Menus with Subheaders and Dividers
   description: Button menus with subheaders, dividers, and compact variant.
 - name: withIcon
   label: Button Menus Using Icons and Menu Alignment
   description: Button menus can display a utility icon next to the dropdown and change the menu alignment.
 - name: menuItems
   label: Button Menu with Styled Items
   description: Button menu items can be disabled, display icons to the left or right of the item label, or both.
 - name: variants
   label: Button Menu Variants
   description: Button menus variants change the border style and size, or display with a dark background.
 - name: onselect
   label: Button Menu with Custom onselect Behavior
   description: Button menu with custom onselect handler.
 - name: iteration
   label: Button Menus Using for:each Iteration for Menu Items
   description: Button menu items can be created from a data source using iteration.
---
A `lightning-button-menu` component represents a button that when clicked displays a
dropdown menu of actions or functions that a user can access.

The menu closes when you click away from it, and it also closes and puts the
focus back on the button when you select a menu item.

Use `lightning-menu-item` as subcomponents of `lightning-button-menu` to specify the menu items for the button menu.

This example shows how to create a dropdown button menu with three items.

```html
<template>
    <lightning-button-menu alternative-text="Settings">
        <lightning-menu-item label="Font" value="font"></lightning-menu-item>
        <lightning-menu-item label="Size" value="size"></lightning-menu-item>
        <lightning-menu-item label="Format" value="format"></lightning-menu-item>
    </lightning-button-menu>
</template>
```

To add a text label to the button before the icon, use the `label` attribute .

The `lightning-button-menu` component supports several variants that change the look of the button. You can use a combination of the `variant`, `icon-name`, and `icon-size` attributes to customize the button and icon styles. 

For more information, see the __Component Styling__ section.

#### Checked Menu Items

You can create menu items that can be checked or unchecked using the `checked`
attribute in the `lightning-menu-item` component, toggling it as needed. To
enable toggling of a menu item, you must set an initial value on the `checked`
attribute, specifying either `true` or `false`.


#### Create Dividers and Subheadings

Use the `lightning-menu-divider` subcomponent to create a dividing line after a menu item. By default, space is added
above and below the divider. Use `variant="compact"` with `lightning-menu-divider` to reduce the space.

Use the `lightning-menu-subheader` subcomponent to create sub-headings in the list of menu items. Specify the text
of the heading using the `label` attribute.

This example shows a dropdown menu with dividers and sub-headings.

```html
 <template>
 <lightning-button-menu label="Menu trigger">
     <lightning-menu-item label="Menu item 1">
     </lightning-menu-item>
     <lightning-menu-divider variant="compact">
         </lightning-menu-divider>
     <lightning-menu-subheader label="Menu sub heading">
         </lightning-menu-subheader>
     <lightning-menu-item label="Menu item 2">
     </lightning-menu-item>
 </lightning-button-menu>
 </template>
```

#### Generate Menu Items

This example creates a button menu with several items during initialization. The `items` array definition uses the `@track` decorator to track mutations in the array. If the value of `items` changes, the component's template rerenders.

```html
<template>

  <lightning-button-menu
        alternative-text="Action"
        onselect={handleMenuSelect}>

    <template for:each={items} for:item="action">
        <lightning-menu-item
            id="action-id"
            label={action.label}
            value={action.value}
            key={action.label}>
        </lightning-menu-item>
    </template>

  </lightning-button-menu>

</template>
```

Define `items` and handle the `select` event in your JavaScript code.

```javascript

    import { LightningElement, track } from 'lwc';

    export default class DemoButtonMenu extends LightningElement {

        @track
        items = [
            {
                id: 'menu-item-1',
                label: 'Alpha',
                value: 'alpha',
            },
            {
                id: 'menu-item-2',
                label: 'Beta',
                value: 'beta',
            },
            {
                id: 'menu-item-3',
                label: 'Gamma',
                value: 'gamma',
            }
        ];



    handleMenuSelect(event) {
        // retrieve the selected item's value
        const selectedItemValue = event.detail.value;

        // INSERT YOUR CODE HERE
    }

}
```
#### Show the Loading State of a Menu

The `is-loading` attribute enables you to show an activity indicator while the menu is loading. You might
use this, for example, to inform users that the menu is working while generating a large list of menu items.
When `is-loading` is `true`, the menu shows a spinner.

Use `loading-state-alternative-text` along with `is-loading` to specify explanatory text such as "Loading menu..." or
"Please wait while items load".

#### Draft Indicators

Use the `is-draft` and `draft-alternative-text` attributes together to indicate that the button menu is in an unsaved state.
The draft indicator, an asterisk, is shown for the button menu when `is-draft` is `true`. The `draft-alternative-text` attribute
is required to provide text describing the reason the menu is considered in a draft state. The button menu draft state might
be used to show there is unsaved state or data that could be lost, for example if there's a user change in a customizable menu.

#### Component Styling

`lightning-button-menu` implements the [menus](https://www.lightningdesignsystem.com/components/menus/) blueprint in the Salesforce Lightning Design System (SLDS).

Use a combination of the `icon-name`, `icon-size`, and `variant` attributes to customize the button and icon styles.

##### Icons

By default, the button displays a utility:down icon to indicate the dropdown function.
Use the `icon-name` attribute to specify an optional [utility icon](https://www.lightningdesignsystem.com/icons/#utility)
to display in front of the utility:down icon. Use the `icon-size` attribute to change the icon size from the default size of `medium`.

When applying SLDS classes or icons, check that they are
available in the SLDS release tied to your org. The Lightning
Design System site shows the latest SLDS resources, and these
become available only when the new release is available in your org.

##### Variants

Use the `variant` attribute with one of these values to apply styling.

* `border` - Shows a gray border around the button's down arrow symbol. The area inside the border is transparent and clickable. This is the default variant and doesn't need to be specified.
* `border-inverse` - Same as the `border` variant except the down arrow is white so it's visible on dark backgrounds.
* `border-filled` - Same as the `border` variant, except the area inside the border is filled with white.
* `container` - Shows only the down arrow, without a visible border. The clickable area surrounding the down arrow is the same size as the default `border` variant and is transparent.
* `bare` - Same as the `container` variant, except the clickable area surrounding the down arrow is smaller than the default.
* `bare-inverse` - Same as the `bare` variant except the down arrow is white so it's visible on dark backgrounds.

##### Sizes

Adjust the button and icon sizes using the `size` attribute with one of these values.

For `bare` and `bare-inverse` variants:
* `medium` is the default size, which creates a 14px by 14px icon
* `small` creates a 12px by 12px icon
* `x-small` creates a 8px by 8px icon
* `large` creates a 24px by 24px icon

For other variants:
* `medium` is the default size, which creates a 32px by 32px button enclosing a 14px by 14px icon.
* `small` creates a 24px by 24px button enclosing a 14px by 14px icon
* `x-small` creates a 20px by 20px button enclosing a 12px by 12px icon
* `xx-small` creates a 16px by 16px button enclosing a 8px by 8px icon

##### Customize Component Styling

To apply additional styling, use the SLDS [utility classes](https://www.lightningdesignsystem.com/utilities/alignment) with the `class` attribute.

This example adds padding around the button menu using an SLDS class.

```html
<lightning-button-menu class="slds-p-around_medium" alternative-text="Show menu">
    <lightning-menu-item value="Edit" label="Edit"></lightning-menu-item>
    <lightning-menu-item value="Save" label="Save"></lightning-menu-item>
</lightning-button-menu>
```

To change the background color of the button, define a custom class using the `class` attribute. This example applies an orange background on the button.

```html
<lightning-button-menu class="my-button-menu" alternative-text="Show menu">
    <lightning-menu-item value="Edit" label="Edit"></lightning-menu-item>
    <lightning-menu-item value="Save" label="Save"></lightning-menu-item>
</lightning-button-menu>
```

Specify the background color on the button using the `sds-c-button-color-background` custom property. Additionally, specify the background color on hover using the `sds-c-button-color-background-hover` custom property.

```css
.my-button-menu {
    --sds-c-button-color-background: orange;
    --sds-c-button-color-background-hover: orange;
}
```

For more information, see [Style Components Using Lightning Design System Styling Hooks](docs/component-library/documentation/lwc/lwc.create_components_css_custom_properties) in the Lightning Web Components Developer Guide.

#### Usage Considerations

Icons are not available in Lightning Out, but they are available in Lightning Components for Visualforce and other experiences.

This component contains menu items that are created only if the button is
triggered. You won't be able to reference the menu items during initialization
or if the button isn't triggered yet.

You can customize the alignment of the dropdown menu relative to the button using `menu-alignment`.
If you are using `button-menu` in a container that specifies the `overflow:hidden` CSS property,
setting `menu-alignment="auto"` ensures that the dropdown menu is not hidden from view when the menu is toggled. For mobile devices, set `menu-alignment="auto"` to ensure proper display of the menu.

When using this component within `lightning-button-group`, set `variant="border-filled"` on `lightning-button-menu` for a white button background. Otherwise, the `lightning-button-menu` background is transparent by default.

#### Accessibility

To inform screen readers that a button is disabled, set the `disabled`
attribute to `true`.

#### Custom Events

**`select`**

The event fired when the menu is selected.

The `select` event returns the following parameter.

Parameter|Type|Description
-----|-----|----------
value|string|The value of the selected option.

The event properties are as follows.

Property|Value|Description
-----|-----|----------
bubbles|false|This event does not bubble.
cancelable|true|This event can be canceled. You can call `preventDefault()` on this event.
composed|false|This event does not propagate outside the template in which it was dispatched.