# lightning-select

> This doc is internal-only.

- [Overview](#overview)
- [Selecting Multiple Options](#selecting-multiple-options)
- [Input Validation](#input-validation)
- [Custom Validity Error Messages](#custom-validity-error-messages)
- [Attributes](#attributes)
- [Methods](#methods)
- [Custom Events](#custom-events)
- [Usage Considerations](#usage-considerations)

## Overview

`lightning-select` enables single and multiple selection on a menu of options using the HTML `select` element and `option` elements. To specify whether multiple options can be selected, use the `multiple` attribute. The `size` attribute can be used to specify how many options are visible at once. `lightning-select` also accepts most general form input attributes such as `required` and `disabled`. Create a menu of options using an array of key-value pair for `label` and `value`, and pass the array to the `options` attribute. 

`lightning-select` implements the [select](https://lightningdesignsystem.com/components/select/) blueprint in the Salesforce Lightning Design System.

Here's an example on how to create a menu of options. Pass a default value to the `value` attribute to make it selected by default when the page loads.

```html
<lightning-select value={selectVal} 
                  label="Select an Org" 
                  options={deptOptions.deptList}
                  onchange={handleChange}></lightning-select>
```

In your JavaScript, define an array of options. Each option has a `value` and `label` property. The `value` is returned when you select an option. The `label` value is the text that's displayed on the menu.

```javascript
import { LightningElement } from 'lwc';

const DEFAULT_VALUE = 'sales';
const DEPT_LIST = [ 
    {
        label: 'Sales',
        value: 'sales',
    },
    {
        label: 'Marketing',
        value: 'marketing',
    },
    {
        label: 'Service',
        value: 'service'
    }
];
export default class SimpleSelect extends LightningElement {
    selectVal = DEFAULT_VALUE;
    deptOptions = {
        deptList: DEPT_LIST,
    };

    handleChange(event) {
        this.selectVal = event.detail.value;
    }
}
```

Selecting an option triggers the `change` event, which calls the
`handleChange` function. To check which option has been clicked,
use `event.detail.value`.

## Selecting Multiple Options

To enable multiple selection of options, use the `multiple` attribute. You can use the `size` attribute to specify the number of options to display by default.

```html
 <lightning-select value={selectVal} 
                   label="Select an Org" 
                   options={deptOptions.deptList}
                   onchange={handleChange}
                   multiple
                   size="3"></lightning-select> 
```

## Input Validation

Client-side input validation is available for this component. You can make
the selection required by adding the `required` attribute. An error message is
automatically displayed when an item is not selected and the element is `required`.

To check the validity states of an input, use the `validity` attribute, which
is based on the `ValidityState` object. You can access the validity states in
your JavaScript. This `validity` attribute returns an object with
`boolean` attributes.

You can override the default message by providing your own value for
`message-when-value-missing`.

Here's an example that displays a custom field-level message when a user interacts with the component but does not select an option.

```html
<lightning-select value={selectVal} 
                  label="Select an Org" 
                  options={deptOptions.deptList}
                  onchange={handleChange}
                  required
                  message-when-value-missing={missingValueMessage}>
</lightning-select> 
```

Set the default value to an empty string so that an option is not selected by default.

```js
import { LightningElement } from 'lwc';

const DEFAULT_VALUE = '';
const DEPT_LIST = [ 
    {
        label: '-- None --',
        value: '',
    },
    {
        label: 'Sales',
        value: 'sales',
    },
    {
        label: 'Marketing',
        value: 'marketing',
    },
    {
        label: 'Service',
        value: 'service'
    }
];
const MISSING_VALUE_MESSAGE = 'Which org are you from?'

export default class RequiredSelect extends LightningElement {
    selectVal = DEFAULT_VALUE;
    deptOptions = {
        deptList: DEPT_LIST,
    };
    missingValueMessage = MISSING_VALUE_MESSAGE;

    handleChange(event) {
        this.selectVal = event.detail.value;
    }
}
```

## Custom Validity Error Messages

`lightning-select` supports `setCustomValidity()` from HTML5's Constraint Validation API.
To set an error message, provide a quoted string to display.
To reset the error message, set the message to an empty string (""). See
details at [https://www.w3.org/TR/html52/sec-forms.html#dom-htmlinputelement-setcustomvalidity](https://www.w3.org/TR/html52/sec-forms.html#dom-htmlinputelement-setcustomvalidity).

This example shows how to display a custom error message with `setCustomValidity()` and `reportValidity()`.

```html
 <lightning-select name="myselect"
                   value={selectVal} 
                   label="Select an Org" 
                   options={deptOptions.deptList}
                   onchange={handleChange}
                   required></lightning-select> 
        
<lightning-button label="Register" onclick={register}></lightning-button>
```

When you click the **Register** button, the `register()` function displays a custom error message if no option is selected, or clears the error message if an option has been selected.

```javascript
import { LightningElement } from 'lwc';

const DEFAULT_VALUE = '';
const DEPT_LIST = [ 
    {
        label: '-- None --',
        value: '',
    },
    {
        label: 'Sales',
        value: 'sales',
    },
    {
        label: 'Marketing',
        value: 'marketing',
    },
    {
        label: 'Service',
        value: 'service'
    }
];
export default class CustomSelectError extends LightningElement {
    selectVal = DEFAULT_VALUE;
    deptOptions = {
        deptList: DEPT_LIST,
    };

    handleChange(event) {
        this.selectVal = event.detail.value;
        console.log(typeof(this.selectVal));
    }

    register(event) {
        const selectCmp = this.template.querySelector("lightning-select");
        if (this.selectVal === '') {
            selectCmp.setCustomValidity("You have not selected an option");
        } else {
            selectCmp.setCustomValidity(""); 
        }
        // Display the error without user interaction
        selectCmp.reportValidity(); 
    }
}

```

## Attributes

Use the following attributes to customize `lightning-select`.

Attribute Name|Type|Description
--------------|----|-----------
accesskey | string | A shortcut key that activates and focuses on the menu.
autocomplete | string | Valid values are `on` (default) and `off`.
class | string | A space-separated list of the CSS classes of the element.
disabled | boolean | Specifies whether the menu is disabled and users cannot interact with it. The default is false.
field-level-help | string | Help text detailing the purpose and function of the menu of options. The text is displayed in a tooltip above the menu.
label | string | The text label for the component. To hide the label but make it available to assistive technologies, use the `label-hidden` variant.
message-when-value-missing | string | The error message that's displayed below the menu when a user interacts with the menu but does not select an option.
multiple | boolean | Specifies whether multiple options can be selected. The default is false.
name | string | The identifier for the component. 
required | boolean | Specifies whether an option must be selected. The default is false.
size | number | The number of rows in the list that should be visible at one time. Use `size` with `multiple`. The default is 4.
value | string | The value of the selected option. If empty and `required` is set to true, the component is in an invalid state.
options | object | An array of menu options with key-value pair for `label` and `value`.
validity | object | Represents the validity states that an element can be in, with respect to constraint validation.
variant | string | The variant changes the appearance of an input field. Accepted variants include standard, `label-inline`, `label-hidden`, and `label-stacked`. This value defaults to standard, which displays the label above the field. `label-hidden` hides the label but make it available to assistive technology. `label-inline` horizontally aligns the label and input field. `label-stacked` places the label above the input field.

## Methods

**`blur()`**

Removes focus from the select element.

**`focus()`**

Sets focus on the select element.

**`checkValidity()`**

Indicates whether the combobox has any validity errors. Returns the valid attribute value (Boolean) on the ValidityState object.

**`reportValidity()`**

Displays the error messages and returns false if the input is invalid. If the input is valid, `reportValidity()` clears displayed error messages and returns true.

**`setCustomValidity(message)`**

Sets a custom error message to be displayed when a form is submitted.
 * `message` (string) - Describes the error. If `message` is an empty string, the error is reset.

## Custom Events

**`change`**

The event fired when an option is selected.

The `change` event returns the following parameter.

Parameter|Type|Description
---------|----|-----------
value|string|The value of the selected option.

The event properties are as follows.

 Property|Value|Description
-----|-----|----------
bubbles|true|This event bubbles up through the DOM.
cancelable|false|This event has no default behavior that can be canceled. You can't call `preventDefault()` on this event.
composed|true|This event propagates outside of the component in which it was dispatched.

## Usage Considerations

`lightning-select` uses `delegatesFocus` to manage focus. `tabindex` is not supported.
See [Handle Focus](https://developer.salesforce.com/docs/component-library/documentation/en/lwc/create_components_focus) for more information.

`lightning-select` has usage differences from its Aura counterpart (`lightning:select`). [`lightning:select`](https://developer.salesforce.com/docs/component-library/bundle/lightning:select/example) does not support multiple selection, and it takes in menu options as subcomponents in `lightning:select`.

`lightning-select` is used as a subcomponent in `lightning-picklist`. Consider `lightning-picklist` if you want a component that renders a combobox for single selection and  a dual listbox for multiple selection. It uses [`lightning-combobox`](https://developer.salesforce.com/docs/component-library/bundle/lightning:combobox/example) and [`lightning-dual-listbox`](https://developer.salesforce.com/docs/component-library/bundle/lightning-dual-listbox/example) as subcomponents. On mobile devices, `lightning-picklist` renders a native select component.

Alternatively, to display dropdown options with icons and autocomplete (typeahead) support, use `lightning-grouped-combobox` and `lightning-mobile-grouped-combobox`.
